@@ -46,15 +46,16 @@ import contextlib
 import sys
 import tempfile
 from glob import glob
-from os import stat
+import os
 from shutil import rmtree
 import textwrap
 import typing
 
-# numpy and rpy2 imports
+import rpy2.rinterface_lib.callbacks
 
 import rpy2.rinterface as ri
-import rpy2.rinterface_lib.callbacks
+
+import rpy2.rinterface_lib.openrlib
 import rpy2.robjects as ro
 import rpy2.robjects.packages as rpacks
 from rpy2.robjects.lib import grdevices
@@ -66,20 +67,21 @@ import warnings
 # Try loading pandas and numpy, emitting a warning if either cannot be
 # loaded.
 try:
-    import numpy  # type: ignore[import]
+    import numpy  # noqa: F401
+    NUMPY_IMPORTED = True
     try:
-        import pandas  # type: ignore[import]
+        import pandas  # noqa: F401
+        PANDAS_IMPORTED = True
     except ImportError as ie:
-        pandas = None
+        PANDAS_IMPORTED = False
         warnings.warn('The Python package `pandas` is strongly '
                       'recommended when using `rpy2.ipython`. '
                       'Unfortunately it could not be loaded '
                       '(error: %s), '
                       'but at least we found `numpy`.' % str(ie))
 except ImportError as ie:
-    # Give up on numerics
-    numpy = None  # type: ignore[assignment]
-    pandas = None
+    NUMPY_IMPORTED = False
+    PANDAS_IMPORTED = False
     warnings.warn('The Python package `pandas` is strongly '
                   'recommended when using `rpy2.ipython`. '
                   'Unfortunately it could not be loaded, '
@@ -88,6 +90,7 @@ except ImportError as ie:
 
 # IPython imports.
 
+import IPython.display  # type: ignore
 from IPython.core import displaypub  # type: ignore
 from IPython.core.magic import (Magics,   # type: ignore
                                 magics_class,
@@ -102,16 +105,27 @@ from IPython.core.magic_arguments import (argument,  # type: ignore
 
 template_converter = get_conversion()
 
+DEVICES_STATIC_RASTER: typing.Set[str] = {'png', 'jpeg'}
+DEVICES_STATIC = DEVICES_STATIC_RASTER | {'svg'}
+DEVICES_SUPPORTED = DEVICES_STATIC | {'X11'}
 
-def _get_ipython_template_converter(template_converter=template_converter,
-                                    numpy=numpy, pandas=pandas):
-    if numpy:
-        from rpy2.robjects import numpy2ri
-        template_converter += numpy2ri.converter
-        if pandas:
+
+if NUMPY_IMPORTED:
+    if PANDAS_IMPORTED:
+        def _get_ipython_template_converter(template_converter=template_converter):
+            from rpy2.robjects import numpy2ri
+            template_converter += numpy2ri.converter
             from rpy2.robjects import pandas2ri
             template_converter += pandas2ri.converter
-    return template_converter
+            return template_converter
+    else:
+        def _get_ipython_template_converter(template_converter=template_converter):
+            from rpy2.robjects import numpy2ri
+            template_converter += numpy2ri.converter
+            return template_converter
+else:
+    def _get_ipython_template_converter(template_converter=template_converter):
+        return template_converter
 
 
 def _get_converter(template_converter=template_converter):
@@ -119,9 +133,7 @@ def _get_converter(template_converter=template_converter):
                      template=template_converter)
 
 
-ipy_template_converter = _get_ipython_template_converter(template_converter,
-                                                         numpy=numpy,
-                                                         pandas=pandas)
+ipy_template_converter = _get_ipython_template_converter(template_converter)
 converter = _get_converter(template_converter=ipy_template_converter)
 
 
@@ -227,6 +239,54 @@ def _parse_input_argument(arg: str) -> typing.Tuple[str, str]:
 #     return res
 
 
+def display_figures(graph_dir, format='png'):
+    """Iterator to display PNG figures generated by R.
+
+    graph_dir : str
+        A directory where R figures are to be searched.
+
+    The iterator yields the image objects."""
+
+    assert format in DEVICES_STATIC_RASTER
+    for imgfile in sorted(glob(os.path.join(graph_dir, f'Rplots*{format}'))):
+        if os.stat(imgfile).st_size >= 1000:
+            img = IPython.display.Image(filename=imgfile)
+            IPython.display.display(img,
+                                    metadata={})
+            # TODO: Synchronization in the console (though it's a bandaid, not a
+            # real solution).
+            sys.stdout.flush()
+            sys.stderr.flush()
+            yield img
+
+
+def display_figures_svg(graph_dir, isolate_svgs=True):
+    """Display SVG figures generated by R.
+
+    graph_dir : str
+        A directory where R figures are to be searched.
+    isolate_svgs : bool
+        Enable SVG namespace isolation in metadata.
+
+    The iterator yields the image object."""
+
+    # as onefile=TRUE, there is only one .svg file
+    imgfile = "%s/Rplot.svg" % graph_dir
+    # Cairo creates an SVG file every time R is called
+    # -- empty ones are not published
+    if os.stat(imgfile).st_size >= 1000:
+        img = IPython.display.SVG(filename=imgfile),
+        IPython.display.display_svg(
+            img,
+            metadata={'image/svg+xml': dict(isolated=True)} if isolate_svgs else {}
+        )
+        # TODO: Synchronization in the console (though it's a bandaid, not a
+        # real solution).
+        sys.stdout.flush()
+        sys.stderr.flush()
+        yield img
+
+
 @magics_class
 class RMagics(Magics):
     """A set of magics useful for interactive work with R via rpy2.
@@ -247,12 +307,12 @@ class RMagics(Magics):
             If True, the published results of the final call to R are
             cached in the variable 'display_cache'.
 
-        device : ['png', 'X11', 'svg']
+        device : ['png', 'jpeg', 'X11', 'svg']
             Device to be used for plotting.
-            Currently only 'png', 'X11' and 'svg' are supported,
-            with 'png' and 'svg' being most useful in the notebook,
-            and 'X11' allowing interactive plots in the terminal.
-
+            Currently only 'png', 'jpeg', 'X11' and 'svg' are supported,
+            with 'X11' allowing interactive plots on a locally-running jupyter,
+            and the other allowing to visualize R figure generated on a remote
+            jupyter server/kernel.
         """
         super(RMagics, self).__init__(shell)
         self.cache_display_data = cache_display_data
@@ -267,17 +327,18 @@ class RMagics(Magics):
         must be installed. Because Cairo forces "onefile=TRUE",
         it is not posible to include multiple plots per cell.
 
-        :param device: ['png', 'X11', 'svg']
-          Device to be used for plotting.
-          Currently only "png" and "X11" are supported,
-          with 'png' and 'svg' being most useful in the notebook,
-          and 'X11' allowing interactive plots in the terminal.
-
+        :param device: ['png', 'jpeg', 'X11', 'svg']
+            Device to be used for plotting.
+            Currently only 'png', 'jpeg', 'X11' and 'svg' are supported,
+            with 'X11' allowing interactive plots on a locally-running jupyter,
+            and the other allowing to visualize R figure generated on a remote
+            jupyter server/kernel.
         """
         device = device.strip()
-        if device not in ['png', 'X11', 'svg']:
+        if device not in DEVICES_SUPPORTED:
             raise ValueError(
-                "device must be one of ['png', 'X11' 'svg'], got '%s'", device)
+                f'device must be one of {DEVICES_SUPPORTED}, got "{device}"'
+            )
         if device == 'svg':
             try:
                 self.cairo = rpacks.importr('Cairo')
@@ -301,8 +362,8 @@ class RMagics(Magics):
     @line_magic
     def Rdevice(self, line):
         """
-        Change the plotting device R uses to one of ['png', 'X11', 'svg'].
-        """
+        Change the plotting device R uses to one of {}.
+        """.format(DEVICES_SUPPORTED)
         self.set_R_plotting_device(line.strip())
 
     def eval(self, code):
@@ -319,21 +380,40 @@ class RMagics(Magics):
         withVisible is a LISPy R function).
         """
         with contextlib.ExitStack() as stack:
+            obj_in_module = (rpy2.rinterface_lib
+                             .callbacks.obj_in_module)
             if self.cache_display_data:
-                stack.enter(
-                    rpy2.rinterface_lib
-                    .callbacks.obj_in_module(rpy2.rinterface_lib.callbacks,
-                                             'consolewrite_print',
-                                             self.write_console_regular)
+                stack.enter_context(
+                    obj_in_module(
+                        rpy2.rinterface_lib.callbacks,
+                        'consolewrite_print',
+                        self.write_console_regular
+                    )
                 )
+            stack.enter_context(
+                obj_in_module(rpy2.rinterface_lib.callbacks,
+                              'consolewrite_warnerror',
+                              self.write_console_regular)
+            )
+            stack.enter_context(
+                obj_in_module(
+                    rpy2.rinterface_lib.callbacks,
+                    '_WRITECONSOLE_EXCEPTION_LOG',
+                    '%s')
+            )
             try:
                 # Need the newline in case the last line in code is a comment.
-                value, visible = ro.r("withVisible({%s\n})" % code)
+                r_expr = ri.parse(code)
+                value, visible = ri.evalr_expr_with_visible(
+                    r_expr
+                )
             except (ri.embedded.RRuntimeError, ValueError) as exception:
                 # Otherwise next return seems to have copy of error.
                 warning_or_other_msg = self.flush()
                 raise RInterpreterError(code, str(exception),
                                         warning_or_other_msg)
+            finally:
+                ro._print_deferred_warnings()
             text_output = self.flush()
             return text_output, value, visible[0]
 
@@ -534,7 +614,7 @@ class RMagics(Magics):
                 args.res = 72
 
         plot_arg_names = ['width', 'height', 'pointsize', 'bg', 'type']
-        if self.device == 'png':
+        if self.device in DEVICES_STATIC_RASTER:
             plot_arg_names += ['units', 'res']
 
         argdict = {}
@@ -544,19 +624,20 @@ class RMagics(Magics):
                 argdict[name] = val
 
         tmpd = None
-        if self.device in ['png', 'svg']:
+        if self.device in DEVICES_STATIC:
             # Create a temporary directory for R graphics output
             # TODO: Do we want to capture file output for other device types
             # other than svg & png?
             tmpd = tempfile.mkdtemp()
             tmpd_fix_slashes = tmpd.replace('\\', '/')
 
-            if self.device == 'png':
-                # Note: that %% is to pass into R for interpolation there
-                grdevices.png("%s/Rplots%%03d.png" % tmpd_fix_slashes,
-                              **argdict)
+            if self.device in DEVICES_STATIC_RASTER:
+                # Note: that %% is to pass into R for interpolation there.
+                rfunc = getattr(grdevices, self.device)
+                rfunc(f'{tmpd_fix_slashes}/Rplots%%03d.{self.device}',
+                      **argdict)
             elif self.device == 'svg':
-                self.cairo.CairoSVG("%s/Rplot.svg" % tmpd_fix_slashes,
+                self.cairo.CairoSVG(f'{tmpd_fix_slashes}/Rplot.svg',
                                     **argdict)
 
         elif self.device == 'X11':
@@ -570,18 +651,22 @@ class RMagics(Magics):
         else:
             # TODO: This isn't actually an R interpreter error...
             raise RInterpreterError(
-                'device must be one of ("png", "X11", "svg")')
+                f'device must be one of {DEVICES_SUPPORTED}')
 
         return tmpd
 
     def publish_graphics(self, graph_dir, isolate_svgs=True):
-        """Wrap graphic file data for presentation in IPython
+        """Wrap graphic file data for presentation in IPython.
+
+        This method is deprecated. Use `display_figures` or
+        'display_figures_svg` instead.
 
         graph_dir : str
             Probably provided by some tmpdir call
         isolate_svgs : bool
             Enable SVG namespace isolation in metadata"""
 
+        warnings.warn('Use method fetch_figures.', DeprecationWarning)
         # read in all the saved image files
         images = []
         display_data = []
@@ -591,7 +676,7 @@ class RMagics(Magics):
 
         if self.device == 'png':
             for imgfile in sorted(glob('%s/Rplots*png' % graph_dir)):
-                if stat(imgfile).st_size >= 1000:
+                if os.stat(imgfile).st_size >= 1000:
                     with open(imgfile, 'rb') as fh_img:
                         images.append(fh_img.read())
         else:
@@ -599,7 +684,7 @@ class RMagics(Magics):
             imgfile = "%s/Rplot.svg" % graph_dir
             # Cairo creates an SVG file every time R is called
             # -- empty ones are not published
-            if stat(imgfile).st_size >= 1000:
+            if os.stat(imgfile).st_size >= 1000:
                 with open(imgfile, 'rb') as fh_img:
                     images.append(fh_img.read().decode())
 
@@ -613,7 +698,7 @@ class RMagics(Magics):
         # Flush text streams before sending figures, helps a little with
         # output.
         for image in images:
-            # Synchronization in the console (though it's a bandaid, not a
+            # TODO: Synchronization in the console (though it's a bandaid, not a
             # real solution).
             sys.stdout.flush()
             sys.stderr.flush()
@@ -850,6 +935,7 @@ class RMagics(Magics):
         tmpd = self.setup_graphics(args)
 
         text_output = ''
+        display_data = []
         try:
             if line_mode:
                 for line in code.split(';'):
@@ -864,14 +950,29 @@ class RMagics(Magics):
                 text_output += text_result
                 if visible:
                     with contextlib.ExitStack() as stack:
+                        obj_in_module = (rpy2.rinterface_lib
+                                         .callbacks
+                                         .obj_in_module)
                         if self.cache_display_data:
                             stack.enter_context(
-                                rpy2.rinterface_lib
-                                .callbacks
-                                .obj_in_module(rpy2.rinterface_lib
-                                               .callbacks,
-                                               'consolewrite_print',
-                                               self.write_console_regular))
+                                obj_in_module(rpy2.rinterface_lib
+                                              .callbacks,
+                                              'consolewrite_print',
+                                              self.write_console_regular)
+                            )
+                        stack.enter_context(
+                            obj_in_module(
+                                rpy2.rinterface_lib.callbacks,
+                                'consolewrite_warnerror',
+                                self.write_console_regular
+                            )
+                        )
+                        stack.enter_context(
+                            obj_in_module(
+                                rpy2.rinterface_lib.callbacks,
+                                '_WRITECONSOLE_EXCEPTION_LOG',
+                                '%s')
+                        )
                         cell_display(result, args)
                         text_output += self.flush()
 
@@ -882,24 +983,25 @@ class RMagics(Magics):
                 print(e.err)
             raise e
         finally:
-            if self.device in ['png', 'svg']:
+            if self.device in DEVICES_STATIC:
                 ro.r('dev.off()')
             if text_output:
                 # display_data.append(('RMagic.R', {'text/plain':text_output}))
                 displaypub.publish_display_data(
-                    data={'text/plain': text_output}, source='RMagic.R')
-            # publish the R images
-            if self.device in ['png', 'svg']:
-                display_data, md = self.publish_graphics(
-                    tmpd, args.isolate_svgs
-                )
-
-                for tag, disp_d in display_data:
-                    displaypub.publish_display_data(data=disp_d, source=tag,
-                                                    metadata=md)
+                    source='RMagic.R',
+                    data={'text/plain': text_output})
+            # publish figures generated by R.
+            if self.device in DEVICES_STATIC_RASTER:
+                for _ in display_figures(tmpd, format=self.device):
+                    if self.cache_display_data:
+                        display_data.append(_)
+            elif self.device == 'svg':
+                for _ in display_figures_svg(tmpd):
+                    if self.cache_display_data:
+                        display_data.append(_)
 
             # kill the temporary directory - currently created only for "svg"
-            # and "png" (else it's None)
+            # and ("png"|"jpeg") (else it's None).
             if tmpd:
                 rmtree(tmpd)
 
